#include "server.h"

#include <sys/time.h>
#include <unistd.h>
#include <string.h>

#define QLEN    5
#define BUFSIZE    4096

#define ERRMSG(str) { printf("ERROR: %s\n", str); }
#define RSHWAIT 20


main(int argc, char **argv)

{
  char   hostaddr[50];
  int    datapsock = 0, datasock = 0;   /* data channel socket descriptor */
  int    ctrlpsock = 0, ctrlsock = 0;   /* control channel socket descriptor */
  int    err_flag = FALSE, pid = 0, pid1 = 0;
  int    res, ret;
  char   filename[80];
  int    status;
  int    north, east;
  int    row_offset, col_offset;
  FILE   *rcfile;
  int    dataport, ctrlport;
  int    n[3], s[3], e[3], w[3];
  int    file_n[3], file_s[3], file_e[3], file_w[3];
  char   **map_filename;
  int    i;

  rcfile = fopen(".2Ddemorc", "r");
  fscanf(rcfile, "%s\n%d\n%d\n", hostaddr, &dataport, &ctrlport);

/* Open data socket */
  printf("Opening data socket \n");
  fflush(stdout);
  if ((datapsock=BindServer(hostaddr, "tcp", "tcp", dataport)) == ERROR) {
    printf("ERROR in binding root port\n");
    exit (ERROR);
  }
/* and control socket */
  printf("Opening control socket \n");
  if ((ctrlpsock=BindServer(hostaddr, "tcp", "tcp", ctrlport)) == ERROR) {
    printf("ERROR in binding control port\n");
    exit (ERROR);
  }

  while (TRUE) {

    if ( (datasock = AcceptClient(datapsock)) == ERROR) {
      err_flag = TRUE;
      break;                   /* parent abort with no child */
    }
    if ( (ctrlsock = AcceptClient(ctrlpsock)) == ERROR) {
      err_flag = TRUE;
	  break;
    }
    if ((pid = fork()) < 0) { 
                /* parent is concurrent server, child does processing */
      perror("server fork error:\n");
      err_flag = TRUE;
	  break;
    } else if (pid == 0) {   /* child */
      close (datapsock);
      close (ctrlpsock);
      break;
    } else {                 /* parent */
      close (datasock);
      close (ctrlsock);
    /* concurrent server loops back to wait on next data connection */  
    }
  }

  if (err_flag) goto cleanup;

  map_filename = (char **) malloc(3*sizeof(char *));
  for (i = 0; i < 3; i++) {
    map_filename[i] = (char *)malloc(80); 
  }
  while (TRUE) {
    res = SendMapBounds(datasock);
    RecvInitCoords(datasock, &north, &east);
    FindFile(north, east, res, &row_offset, &col_offset,
	         filename);
    if ((pid1 = fork()) < 0) {
      perror("fork error:");
      break;
    } else if (pid1 == 0) { /* child sends the urls */
      SendUrlList(ctrlsock);
      break;            /* and exits */
    } 
    /* parent executes from here */
    SendInitSegments(datasock, row_offset, col_offset, filename);
    while (TRUE) {
      if ((ret = RecvNewCoords(datasock, n, s, e, w, file_n, file_s,
                               file_e, file_w, map_filename)) == -1) break;
      ret = SendNewSegments(datasock, res, n, s, e, w, file_n,
                            file_s, file_e, file_w, map_filename);
      if (ret == -1) break;
    }
    break;
  }
  
cleanup:
/* close all socket streams */
  if (datapsock)
    close (datapsock);
  if (ctrlpsock)
    close (ctrlpsock);
  if (datasock)
    close (datasock);
  if (ctrlsock)
    close (ctrlsock);
  if (rcfile) fclose (rcfile);
  for (i = 0; i < 3; i++) {
    if (map_filename[i]) free(map_filename[i]);
  }
  if (map_filename) free(map_filename);

}





/* ----------------------------------------------------------------------
Name: int RecvMapBounds(int *news)
In: None
Out: sockfd
Comments: This function connects the data and control sockets to the
          server and returns the datasockfd and ctrlsockfd
------------------------------------------------------------------------- */
int SendMapBounds(int datasock) {

  FILE *fd;
  int  north, south, east, west, res;
  char buf[100];

  fd = fopen("bounds", "r");
  if (!fd) {
    perror("error: could not open file: bounds: \n");
    return(ERROR);
  }
  fscanf(fd, "%d %d %d %d\n", &north, &south, &east, &west);
  fscanf(fd, "%d", &res);
  if (fd) fclose(fd);
  sprintf(buf, "%d %d %d %d", north, south, east, west);
  if (WriteBytes(datasock, buf, 100) != 100) {
      printf("ERROR: In writing map boundaries to data channel\n");
      return (ERROR);
  }
  if (WriteInt(datasock, &res, 1) != 1) {
      printf("ERROR: In writing map resolution data to data channel\n");
      return (ERROR);
  }

  return (res);
}




/* ----------------------------------------------------------------------
Name: int RecvMapBounds(int *news)
In: None
Out: sockfd
Comments: This function connects the data and control sockets to the
          server and returns the datasockfd and ctrlsockfd
------------------------------------------------------------------------- */
int RecvInitCoords(int datasock, int *north, int *east) {
  
  char buf[120];

  if (ReadBytes(datasock, buf, 120) != 120) {
    printf("ERROR: In reading initial coords from data channel\n");
    return (ERROR);
  }

  sscanf(buf, "%d %d", north, east);
  return (OK);
}





/* ----------------------------------------------------------------------
Name: int RecvMapBounds(int *news)
In: None
Out: sockfd
Comments: This function connects the data and control sockets to the
          server and returns the datasockfd and ctrlsockfd
------------------------------------------------------------------------- */
int FindFile(int north, int east, int res, int *row_offset, 
             int *col_offset, char *f_name) {

  FILE *fd;
  int  f_north, f_south, f_east, f_west;
  int  temp1, temp2, temp3, temp4;
  
  fd = fopen("file_bounds", "r");
  if (!fd) {
    perror("error: could not open file: file_bounds: \n");
    return(ERROR);
  }

  /* first find the file in which the map segments lie */
  while (TRUE) {
    memset(f_name, '\0', 50);
    if (fscanf(fd, "%d %d %d %d %s\n", &f_north, &f_south, &f_east, &f_west,
	   f_name) == EOF) {
      perror("Could not find initial segment in the map files:");
      return (ERROR);
    }
    if (north <= f_north && north >= f_south && east >= f_west
	&& east <= f_east)
      break;
  }
    if (fd) fclose (fd);

  /* and then compute the file row/column offsets for the northwest corner*/
printf("east: %d f_west: %d res: %d\n", east, f_west, res);
  temp1 = (int)floor((double)((east - f_west)/res));
printf("temp1: %d\n", temp1);
  temp2 = (int)floor((double)((f_north - north)/res));
  temp3 = (int)floor((double)(temp1/SEGMENT_WIDTH));
printf("temp3: %d\n", temp3);
  temp4 = (int)floor((double)(temp2/SEGMENT_HEIGHT));
  *row_offset = temp4 * SEGMENT_HEIGHT;
  *col_offset = temp3 * SEGMENT_WIDTH;

  return(OK);

}





/* ----------------------------------------------------------------------
Name: int RecvMapBounds(int *news)
In: None
Out: sockfd
Comments: This function connects the data and control sockets to the
          server and returns the datasockfd and ctrlsockfd
------------------------------------------------------------------------- */
int SendUrlList(int ctrlsock) {

  FILE *fdloc;
  int  n, e;
  char location[80];
  char url[120];
  char buf[250];
  int  status, count;

  fdloc = fopen("loclist", "r");
  if (!fdloc) {
    perror("error: could not open file: loclist: \n");
    return(ERROR);
  }

  fscanf(fdloc, "%d\n", &count);
  if (WriteInt(ctrlsock, &count, 1) != 1) {
    printf("ERROR: In writing url count to control channel\n");
    if (fdloc) fclose (fdloc);
    return (ERROR);
  } 
  while (TRUE) {
    memset(location, '\0', 80);
    memset(url, '\0', 120);
    status = fscanf(fdloc, "%d %d %s %s\n", &n, &e, location, url);
    memset(buf, '\0', 250);
    if (status == EOF)
      sprintf(buf, "%d", -1);
    else
      sprintf(buf, "%d %d %s %s", n, e, location, url);
    if (WriteBytes(ctrlsock, buf, 250) != 250) {
      printf("ERROR: In writing urls to data channel\n");
      return (ERROR);
    } 
    if (status == EOF) break;
  }
  if (fdloc) fclose (fdloc);

  return (OK);

}





/* ----------------------------------------------------------------------
Name: int RecvMapBounds(int *news)
In: None
Out: sockfd
Comments: This function connects the data and control sockets to the
          server and returns the datasockfd and ctrlsockfd
NOTE: Boundary conditions not being taken care of!! Need to modify this 
	  funciton to take care of the special cases at boudaries.
------------------------------------------------------------------------- */
int SendInitSegments(int datasock, int row_offset, int col_offset,
		             char *filename) {

  char *buf[3], *line;
  FILE *fd;
  int  fptr;
  int  size, i, j, line_size, l;
  int  roffset, coffset, file_offset, buf0_ptr, buf1_ptr, buf2_ptr;
  int  line_ptr;

  size = SEGMENT_WIDTH * SEGMENT_HEIGHT * 3;
  for (i = 0; i < 3; i++) 
    buf[i] = (char *)malloc(size);
  line_size = NUM_SEGMENTS_IN_ROW*SEGMENT_WIDTH*3;
  line = (char *)malloc(line_size);
  
  fd = fopen(filename, "r");
  if (!fd) {
    perror("error: could not open data file for initial coords: \n");
    return(ERROR);
  }

  for (l = 0; l < 3; l++) {
    if (l == 0)
      roffset = row_offset - SEGMENT_HEIGHT;
    else if (l == 1)
      roffset = row_offset;
    else
      roffset = row_offset + SEGMENT_HEIGHT;
    if (roffset == 0)
      file_offset = 0;
    else 
      file_offset = (roffset-1) * SEGMENT_WIDTH * NUM_SEGMENTS_IN_ROW * 3; 
                               /* n segments per row and 3 bytes per pixel */
    coffset = col_offset - SEGMENT_WIDTH;  
    file_offset += coffset * 3;
    buf0_ptr = 0;
    buf1_ptr = 0;
    buf2_ptr = 0;
    for (j = 0; j < SEGMENT_HEIGHT; j++) {
      fseek(fd, file_offset, 0); 
                        /* seek to the beg. of the line from where to read */
      fread(line, line_size, 1, fd); /* read in the line from the map file */
      if (roffset == 0)
        file_offset = (j+1) * SEGMENT_WIDTH * NUM_SEGMENTS_IN_ROW * 3; 
      else
        file_offset = (roffset+j) * SEGMENT_WIDTH * NUM_SEGMENTS_IN_ROW * 3;
      file_offset += coffset * 3;
      line_ptr = 0;
      for (i = 0; i < 3*SEGMENT_WIDTH; i++) {
        if (i < SEGMENT_WIDTH) {
          buf[0][buf0_ptr++] = line[line_ptr++];
          buf[0][buf0_ptr++] = line[line_ptr++];
          buf[0][buf0_ptr++] = line[line_ptr++];
        } else if (i < 2*SEGMENT_WIDTH) {
          buf[1][buf1_ptr++] = line[line_ptr++];
          buf[1][buf1_ptr++] = line[line_ptr++]; 
          buf[1][buf1_ptr++] = line[line_ptr++]; 
        } else {
          buf[2][buf2_ptr++] = line[line_ptr++];
          buf[2][buf2_ptr++] = line[line_ptr++]; 
          buf[2][buf2_ptr++] = line[line_ptr++]; 
        }
      }
    }
 /* the three buffers are ready, so send them now */
    for (i = 0; i < 3 ; i++) {
      if (WriteBytes(datasock, buf[i], size) != size) {
        printf("ERROR: In writing segment buffer to data channel\n");
        return (ERROR);
      }
    }
  }

  if (fd) fclose(fd);
  if (line) free(line);
  for (i = 0; i < 3; i++)
    if (buf[i]) free (buf[i]);

  return(OK);
}




/* ----------------------------------------------------------------------
Name: int RecvMapBounds(int *news)
In: None
Out: sockfd
Comments: This function connects the data and control sockets to the
          server and returns the datasockfd and ctrlsockfd
NOTE: Boundary conditions not being taken care of!! Need to modify this
      funciton to take care of the special cases at boudaries.
------------------------------------------------------------------------- */
int RecvNewCoords(int datasock, int *north, int *south, int *east,
                  int *west, int *file_n, int *file_s, int *file_e,
                  int *file_w, char **filename) {

  char msgbuf[120];
  int  seg0_north, seg0_west, seg0_south, seg0_east;
  int  seg1_north, seg1_west, seg1_south, seg1_east;
  int  seg2_north, seg2_west, seg2_south, seg2_east;
  int  s;
  FILE *fd;

  if (ReadBytes(datasock, msgbuf, 120) != 120) {
    printf("ERROR: In reading new segment coords from data channel\n");
    return (ERROR);
  }
  sscanf(msgbuf, "%d %d %d %d %d %d %d %d %d %d %d %d",
         &seg0_north, &seg0_west, &seg0_south, &seg0_east,
         &seg1_north, &seg1_west, &seg1_south, &seg1_east,
         &seg2_north, &seg2_west, &seg2_south, &seg2_east);
  printf("server got: %s\n", msgbuf);
  if (seg0_north == -1) {
    printf("Goodbye\n");
    return (-1);  /* shutdown */
  }

  /* get the data for each of the requested segments */
  fd = fopen("file_bounds", "r");
  if (!fd) {
    perror("error: could not open file: file_bounds: \n");
    return(ERROR);
  }

  for (s = 0; s < 3; s++) {
    if (s == 0) {
      north[s] = seg0_north;
      south[s] = seg0_south;
      east[s] = seg0_east;
      west[s] = seg0_west;
    } else if (s == 1) {
      north[s] = seg1_north;
      south[s] = seg1_south;
      east[s] = seg1_east;
      west[s] = seg1_west;
    } else {
      north[s] = seg2_north;
      south[s] = seg2_south;
      east[s] = seg2_east;
      west[s] = seg2_west;
    }
    rewind(fd);
    while (TRUE) { /* first find the file */
      memset(filename[s], '\0', 80);
      if (fscanf(fd, "%d %d %d %d %s\n", &file_n[s], &file_s[s], &file_e[s],
                 &file_w[s], filename[s]) == EOF) {
        perror("Could not find requested segment in the map files:");
        if (fd) fclose (fd);
        return (ERROR);
      }
      if ((north[s] <= file_n[s] && south[s] >= file_s[s]) &&
          (west[s] >= file_w[s] && east[s] <= file_e[s])) {  /* found it! */
        break;
      }
    }
  }

printf("files are: %s %s %s\n", filename[0], filename[1], filename[2]);
  if (fd) fclose (fd);
  return (OK);
}




/* ----------------------------------------------------------------------
Name: int RecvMapBounds(int *news)
In: None
Out: sockfd
Comments: This function connects the data and control sockets to the
          server and returns the datasockfd and ctrlsockfd
NOTE: Boundary conditions not being taken care of!! Need to modify this
      funciton to take care of the special cases at boudaries.
------------------------------------------------------------------------- */
int SendNewSegments(int datasock, int res, int *north, int *south, 
                    int *east, int *west, int *file_n, int *file_s,
                    int *file_e, int *file_w, char **filename) {

  char *buf, *line;
  FILE *fd, *fdmap;
  int  temp1, temp2, temp3, temp4, row_offset, col_offset, file_offset;
  int  size, buf_ptr, line_ptr, line_size;
  int  i, j, s;
  struct timeval *tp;
  struct timezone *tzp;

/* timer structures */
  tp = (struct timeval *)malloc(sizeof(struct timeval));
  tzp = (struct timezone *)malloc(sizeof(struct timezone));

  size = SEGMENT_WIDTH * SEGMENT_HEIGHT * 3;
  buf = (char *)malloc(size * sizeof(char));
  line_size = SEGMENT_WIDTH * 3;
  line = (char *)malloc(line_size * sizeof(char));

  for (s = 0; s < 3; s++) {
    /* and then the offset for the data in the file */
    temp1 = (int)floor((double)((west[s] - file_w[s])/res)); /* nw corner */
    temp2 = (int)floor((double)((file_n[s] - north[s])/res));
    temp3 = (int)floor((double)(temp1/SEGMENT_WIDTH));
    temp4 = (int)floor((double)(temp2/SEGMENT_HEIGHT));
    row_offset = temp4 * SEGMENT_HEIGHT;
    col_offset = temp3 * SEGMENT_WIDTH;
    /* now read the data */
    fdmap = fopen(filename[s], "r");
    if (row_offset == 0)
      file_offset = 0;
    else
      file_offset = (row_offset-1) * SEGMENT_WIDTH * NUM_SEGMENTS_IN_ROW * 3;
                    /* n segments per row and 3 bytes per pixel */
    file_offset += col_offset * 3;
    buf_ptr = 0;
    for (j = 0; j < SEGMENT_HEIGHT; j++) {
      fseek(fdmap, file_offset, 0);
                        /* seek to the beg. of the line from where to read */
      fread(line, line_size, 1, fdmap); 
                        /* read a line from the map file */
      if (row_offset == 0)
        file_offset = (j+1) * SEGMENT_WIDTH * NUM_SEGMENTS_IN_ROW * 3;
      else
        file_offset = (row_offset+j) * SEGMENT_WIDTH * NUM_SEGMENTS_IN_ROW * 3;
      file_offset += col_offset * 3;
      line_ptr = 0;
      for (i = 0; i < SEGMENT_WIDTH; i++) {
        buf[buf_ptr++] = line[line_ptr++];
        buf[buf_ptr++] = line[line_ptr++];
        buf[buf_ptr++] = line[line_ptr++];
      }
    }
    if (fdmap) fclose (fdmap);
    /* and send it */
    if (WriteBytes(datasock, buf, size) != size) {
      printf("ERROR: In writing segment buffer to data channel\n");
      return (ERROR);
    }
  }

  return (OK);
}

